from jitcode import jitcode
from symengine import Symbol
from itertools import permutations

p = Symbol('p')

f = [1/p]

for integrator in ["dopri5","RK45"]:
	def set_integrator(ODE):
		ODE.set_integrator(integrator)
	
	def set_parameters(ODE):
		ODE.set_parameters(1)
	
	def set_initial_value(ODE):
		ODE.set_initial_value([0], 0)
	
	at_east_one_working_order = False
	for functions in permutations([set_integrator,set_parameters,set_initial_value]):
		ODE = jitcode(f, control_pars=[p], verbose=False)
		try:
			for function in functions:
				function(ODE)
		except RuntimeError as exception:
			assert str(exception).startswith("Something needs parameters to be set")
		else:
			assert abs(ODE.integrate(1.)-1) < 1e-8
			at_east_one_working_order = True
		
		print( ".", end="", flush=True )
	assert at_east_one_working_order

f_2 = [1]

for integrator in ["dopri5","RK45"]:
	def set_integrator(ODE):
		ODE.set_integrator(integrator)
	
	def set_initial_value(ODE):
		ODE.set_initial_value([0], 0)
	
	at_east_one_working_order = False
	for functions in permutations([set_integrator,set_initial_value]):
		ODE = jitcode(f_2, verbose=False)
		try:
			for function in functions:
				function(ODE)
		except RuntimeError as exception:
			assert str(exception).startswith("Something needs parameters to be set")
		else:
			assert abs(ODE.integrate(1.)-1) < 1e-8
			at_east_one_working_order = True
		
		print( ".", end="", flush=True )
	assert at_east_one_working_order

print("")



import sys
import os
from unittest.mock import MagicMock as Mock
from setuptools_scm import get_version

MOCK_MODULES = [
	'numpy', 'numpy.testing', 'numpy.random',
	'scipy', 'scipy.integrate', 'scipy.integrate._ode', 'scipy.stats', 'scipy.integrate._ivp.ivp',
	'symengine', 'symengine.printing', 'symengine.lib.symengine_wrapper',
	'jitcxde_common.helpers','jitcxde_common.numerical','jitcxde_common.symbolic','jitcxde_common.transversal',
	]
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

class ode_mock(object): pass
sys.modules['scipy.integrate'] = Mock(ode=ode_mock)
class GroupHandler_mock(object): pass
sys.modules['jitcxde_common.transversal'] = Mock(GroupHandler=GroupHandler_mock)

sys.path.insert(0,os.path.abspath("../examples"))
sys.path.insert(0,os.path.abspath("../jitcode"))

needs_sphinx = '1.3'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'numpydoc',
    'sphinx.ext.graphviz'
]

source_suffix = '.rst'

master_doc = 'index'

project = u'JiTCODE'
copyright = u'2016, Gerrit Ansmann'

release = version = get_version(root='..', relative_to=__file__)

default_role = "any"

add_function_parentheses = True

add_module_names = False

html_theme = 'nature'
pygments_style = 'colorful'
htmlhelp_basename = 'JiTCODEdoc'

numpydoc_show_class_members = False
autodoc_member_order = 'bysource'

graphviz_output_format = "svg"

def on_missing_reference(app, env, node, contnode):
	if node['reftype'] == 'any':
		return contnode
	else:
		return None

def setup(app):
	app.connect('missing-reference', on_missing_reference)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import unittest

import numpy as np
from numpy.testing import assert_allclose
from scipy.stats import sem as standard_error
from symengine import symbols

from jitcode import jitcode, y, jitcode_lyap, UnsuccessfulIntegration, test
from jitcode._jitcode import _is_C, _is_lambda

from scenarios import (
		y0, f_of_y0, jac_of_y0, y1, lyaps, vanilla, n,
	)

class TestOrders(unittest.TestCase):
	"""
	tests that the derivative and Jacobian are compiled/generated as intended when calling the several methods of the jitcode object in a certain order.
	"""
	def test_standard_order(self):
		self.ODE = jitcode(**vanilla)
		self.ODE.set_integrator("dopri5")
		self.ODE.set_initial_value(y0)
		self.assertTrue(_is_C(self.ODE.f))
		self.assertIsNone(self.ODE.jac)
	
	def test_standard_order_with_jac(self):
		self.ODE = jitcode(**vanilla)
		self.ODE.set_integrator("lsoda")
		self.ODE.set_initial_value(y0)
		self.assertTrue(_is_C(self.ODE.f))
		self.assertTrue(_is_C(self.ODE.jac))
	
	def test_initialise_first(self):
		self.ODE = jitcode(**vanilla)
		self.ODE.set_initial_value(y0)
		self.ODE.set_integrator("dop853")
		self.assertTrue(_is_C(self.ODE.f))
	
	def test_initalise_with_dict(self):
		self.ODE = jitcode(**vanilla)
		initial_value = {y(i):y0[i] for i in range(n)}
		self.ODE.set_initial_value(initial_value)
		self.ODE.set_integrator("dop853")
		self.assertTrue(_is_C(self.ODE.f))
	
	def test_lambdas_with_jac(self):
		self.ODE = jitcode(wants_jacobian=True,**vanilla)
		self.ODE.generate_lambdas()
		self.ODE.set_integrator("lsoda")
		self.ODE.set_initial_value(y0)
		self.assertTrue(_is_lambda(self.ODE.f))
		self.assertTrue(_is_lambda(self.ODE.jac))
	
	def test_lambdas_without_jac(self):
		self.ODE = jitcode(**vanilla)
		self.ODE.generate_lambdas()
		self.assertIsNone(self.ODE.jac)
		self.ODE.set_integrator("vode")
		self.ODE.set_initial_value(y0)
	
	def test_compile_without_jac(self):
		self.ODE = jitcode(**vanilla)
		self.ODE.compile_C()
		self.ODE.set_integrator("LSODA")
		self.ODE.set_initial_value(y0)
		self.assertTrue(_is_C(self.ODE.f))
		self.assertTrue(_is_C(self.ODE.jac))
	
	def test_generate_jac_manually(self):
		self.ODE = jitcode(**vanilla)
		self.ODE.generate_jac_C()
		self.ODE.set_integrator("vode")
		self.ODE.set_initial_value(y0)
		self.assertTrue(_is_C(self.ODE.f))
		self.assertTrue(_is_C(self.ODE.jac))
	
	def test_save_and_load(self):
		self.ODE = jitcode(**vanilla)
		filename = self.ODE.save_compiled(overwrite=True)
		self.ODE = jitcode(n=n,module_location=filename)
		self.ODE.set_integrator("RK45")
		self.ODE.set_initial_value(y0)
		self.assertTrue(_is_C(self.ODE.f))
		self.assertIsNone(self.ODE.jac)
	
	def tearDown(self):
		if isinstance(self.ODE,jitcode) and self.ODE.f_sym():
			self.ODE.check()
		self.assertIsNotNone(self.ODE.f)
		assert_allclose( self.ODE.f(0.0,y0), f_of_y0, rtol=1e-5 )
		if not self.ODE.jac is None:
			assert_allclose( self.ODE.jac(0.0,y0), jac_of_y0, rtol=1e-5)
		assert_allclose( self.ODE.integrate(1.0), y1, rtol=1e-4 )
		for i in reversed(range(n)):
			assert_allclose( self.ODE.y_dict[y(i)], y1[i], rtol=1e-4 )

integrators = [
		#  name     with_jac
		("dopri5"  , False),
		("dop853"  , False),
		("lsoda"   , True ),
		("vode"    , True ),
		("RK23"    , False),
		("RK45"    , False),
		("Radau"   , True ),
		("BDF"     , True ),
		("LSODA"   , True ),
	]

class TestIntegrators(unittest.TestCase):
	"""
		Tests for every known integrator that its properties are identified correctly and the derivative and Jacobian are generated as needed.
	"""
	def test_normal_integration(self):
		for lambdas in (True,False):
			for integrator in integrators:
				with self.subTest(integrator=integrator,lambdas=lambdas):
					ODE = jitcode(**vanilla)
					if lambdas:
						ODE.generate_f_lambda()
						ODE.generate_jac_lambda()
					ODE.set_integrator(integrator[0])
					assert ODE._wants_jacobian == integrator[1]
					assert not  _is_C(ODE.f) == lambdas
					assert _is_lambda(ODE.f) == lambdas
					if integrator[1]:
						assert _is_lambda(ODE.jac) == lambdas
						assert not  _is_C(ODE.jac) == lambdas
					ODE.set_initial_value(y0,0.0)
					assert_allclose( ODE.integrate(1.0), y1, rtol=1e-3 )

class TestLyapunov(unittest.TestCase):
	"""
		Integration test for jitcode_lyap.
	"""
	def test_regular(self):
		self.ODE = jitcode_lyap(**vanilla,n_lyap=n)
		self.ODE.set_integrator("dopri5")
	
	def test_jac(self):
		self.ODE = jitcode_lyap(**vanilla,n_lyap=n)
		self.ODE.set_integrator("lsoda")
	
	def test_save_and_load_with_jac(self):
		self.ODE = jitcode_lyap(**vanilla,n_lyap=n,wants_jacobian=True)
		filename = self.ODE.save_compiled(overwrite=True)
		self.ODE = jitcode_lyap((),n=n,n_lyap=n,module_location=filename)
		self.ODE.set_integrator("vode")
		self.assertTrue(_is_C(self.ODE.f))
		self.assertTrue(_is_C(self.ODE.jac))
	
	def tearDown(self):
		if isinstance(self.ODE,jitcode) and self.ODE.f_sym():
			self.ODE.check()
		self.ODE.set_initial_value(y0,0.0)
		times = range(10,100000,10)
		data = np.vstack([ self.ODE.integrate(time)[1] for time in times ])
		result = np.average(data[1000:], axis=0)
		margin = standard_error(data[1000:], axis=0)
		self.assertLess( np.max(margin), 0.003 )
		for i in range(n):
			self.assertLess( result[i]-lyaps[i], 5*margin[i] )

class TestErrors(unittest.TestCase):
	def test_duplicate_error(self):
		ODE1 = jitcode(**vanilla)
		ODE2 = jitcode(**vanilla)
		ODE1.compile_C(modulename="foo")
		with self.assertRaises(NameError):
			ODE2.compile_C(modulename="foo")
	
	def test_wrong_n(self):
		with self.assertRaises(ValueError):
			ODE = jitcode(**vanilla,n=2*n)
	
	def test_dimension_mismatch(self):
		ODE = jitcode(**vanilla)
		with self.assertRaises(ValueError):
			ODE.set_initial_value(np.array([1.,2.,3.]),0.0)
	
	def test_check_index_negative(self):
		ODE = jitcode([y(-1)])
		with self.assertRaises(ValueError):
			ODE.check()
	
	def test_check_index_too_high(self):
		ODE = jitcode([y(1)])
		with self.assertRaises(ValueError):
			ODE.check()
	
	def test_check_undefined_variable(self):
		x = symbols("x")
		ODE = jitcode([x])
		with self.assertRaises(ValueError):
			ODE.check()
	
	def test_no_integrator(self):
		ODE = jitcode(**vanilla)
		with self.assertRaises(RuntimeError):
			ODE.integrate(1.0)
	
	def test_no_interpolation_LSODA(self):
		ODE = jitcode(**vanilla)
		with self.assertRaises(NotImplementedError):
			ODE.set_integrator("LSODA",interpolate=False)

if __name__ == "__main__":
	unittest.main(buffer=True)




from numpy import pi, vstack, savetxt
from jitcode import jitcode, y
from symengine import sin
from numpy.random import choice, uniform

n = 100
c = 3.0
q = 0.2

A = choice( [1,0], size=(n,n), p=[q,1-q] )

omega = uniform(-0.5,0.5,n)
# Sorting does not effect the qualitative dynamics, but only the resulting plots:
omega.sort()

def kuramotos_f():
	for i in range(n):
		coupling_sum = sum(
				sin(y(j)-y(i))
				for j in range(n)
				if A[j,i]
			)
		yield omega[i] + c/(n-1)*coupling_sum

I = jitcode(kuramotos_f,n=n)
I.set_integrator("dop853",atol=1e-6,rtol=0)

initial_state = uniform(0,2*pi,n)
I.set_initial_value(initial_state,time=0.0)

times = range(0,2001)
data = vstack ([ I.integrate(time) for time in times ])
data %= 2*pi
savetxt("kuramotos.dat",data)




import numpy as np
from numpy.random import choice, uniform
from time import process_time
from scipy.integrate import ode, solve_ivp, odeint
from scipy.integrate._ivp.ivp import METHODS
from jitcode import jitcode, y
from symengine import sin

solver_ode = "dopri5"
solver_ivp = "RK45"

# Context manager for timing
class timer(object):
	def __init__(self,name):
		self.name = name
	
	def __enter__(self):
		self.start = process_time()
	
	def __exit__(self,*args):
		end = process_time()
		duration = end-self.start
		print("%s took %.5f s" % (self.name,duration))

# The actual test
def test_scenario(name,fun,initial,times,rtol,atol):
	print(40*"-",name,40*"-",sep="\n")
	
	with timer("ode (%s)"%solver_ode):
		I = ode(fun)
		I.set_integrator(solver_ode,rtol=rtol,atol=atol,nsteps=10**8)
		I.set_initial_value(initial,0.0)
		result = np.vstack([I.integrate(time) for time in times])
	assert I.successful()
	
	inv_fun = lambda y,t: fun(t,y)
	with timer("odeint with suboptimal function (LSODA)"):
		result = odeint(
				func=inv_fun,
				y0=initial, t=[0.0]+list(times),
				rtol=rtol, atol=atol,
				mxstep=10**8
				)
	
	with timer("solve_ivp (%s) without result"%solver_ivp):
		I = solve_ivp(
				fun,
				t_span=(0,times[-1]),
				y0=initial,
				method=solver_ivp, rtol=rtol, atol=atol
			)
	assert I.status != -1
	
	with timer("solve_ivp (%s)"%solver_ivp):
		I = solve_ivp(
				fun,
				t_span=(0,times[-1]), t_eval=times,
				y0=initial,
				method=solver_ivp, rtol=rtol, atol=atol
			)
		result = I.y
	assert I.status != -1
	
	with timer("solve_ivp (%s) with dense_output"%solver_ivp):
		I = solve_ivp(
				fun,
				t_span=(0,times[-1]),
				y0=initial,
				method=solver_ivp, rtol=rtol, atol=atol,
				dense_output=True
			)
		result = np.vstack([I.sol(time) for time in times])
	assert I.status != -1
	
	with timer("%s with dense output"%solver_ivp):
		I = METHODS[solver_ivp](
				fun=fun,
				y0=initial, t0=0.0, t_bound=times[-1],
				rtol=rtol, atol=atol
			)
		def solutions():
			for time in times:
				while I.t < time:
					I.step()
				yield I.dense_output()(time)
		result = np.vstack(list(solutions()))
	assert I.status != "failed"
	
	with timer("%s with manual resetting"%solver_ivp):
		I = METHODS[solver_ivp](
				fun=fun,
				y0=initial, t0=0.0, t_bound=times[-1],
				rtol=rtol, atol=atol
			)
		def solutions():
			for time in times:
				I.t_bound = time
				I.status = "running"
				while I.status == "running":
					I.step()
				yield I.y
		result = np.vstack(list(solutions()))
	assert I.status != "failed"
	
	with timer("%s with reinitialising"%solver_ivp):
		def solutions():
			current_time = 0.0
			state = initial
			for time in times:
				I = METHODS[solver_ivp](
						fun=fun,
						y0=state, t0=current_time, t_bound=time,
						rtol=rtol, atol=atol
					)
				while I.status == "running":
					I.step()
				assert I.status != "failed"
				current_time = time
				state = I.y
				yield state
		result = np.vstack(list(solutions()))

# Using compiled functions to make things faster
def get_compiled_function(f):
	dummy = jitcode(f,verbose=False)
	dummy.compile_C()
	return dummy.f

# The actual scenarios
test_scenario(
	name = "two coupled FitzHugh–Nagumo oscillators",
	fun = get_compiled_function([
			y(0)*(-0.025794-y(0))*(y(0)-1.0)-y(1)+0.128*(y(2)-y(0)),
			0.0065*y(0)-0.02*y(1),
			y(2)*(-0.025794-y(2))*(y(2)-1.0)-y(3)+0.128*(y(0)-y(2)),
			0.0135*y(2)-0.02*y(3)
		]),
	initial = np.array([1.,2.,3.,4.]),
	times = 2000+np.arange(0,100000,10),
	rtol = 1e-5,
	atol = 1e-8,
)


n, c, q = 100, 3.0, 0.2
A = choice( [1,0], size=(n,n), p=[q,1-q] )
omega = uniform(-0.5,0.5,n)
def kuramotos_f():
	for i in range(n):
		coupling_sum = sum(
				sin(y(j)-y(i))
				for j in range(n)
				if A[j,i]
			)
		yield omega[i] + c/(n-1)*coupling_sum

test_scenario(
	name = "random network of Kuramoto oscillators",
	fun = get_compiled_function(kuramotos_f),
	initial = uniform(0,2*np.pi,n),
	times = range(1,10001,10),
	rtol = 1e-13,
	atol = 1e-6,
)





#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Suppose, we want to implement the Lotka–Volterra model, which is described by the following equations:

.. math::
	
	\\begin{alignat*}{3}
	\\dot{B} &=&    γ · B &- φ · R · B\\\\
	\\dot{R} &=&\\, -ω · R &+ ν · R · B
	\\end{alignat*}

with :math:`γ = 0.6`, :math:`φ = 1.0`, :math:`ω = 0.5`, and :math:`ν = 0.5`.

We start with a few imports:

.. literalinclude:: ../examples/lotka_volterra.py
	:start-after: example-st\u0061rt
	:lines: 1-2
	:dedent: 1

… and defining the control parameters:

.. literalinclude:: ../examples/lotka_volterra.py
	:start-after: example-st\u0061rt
	:lines: 4-7
	:dedent: 1

The `y` that we imported from `jitcode` has to be used for the dynamical variables. However, to make our code use the same notation as the above equation, we can rename them:

.. literalinclude:: ../examples/lotka_volterra.py
	:start-after: example-st\u0061rt
	:lines: 9
	:dedent: 1

We here might as well have written `R,B = y(0),y(1)`, but the above is more handy for larger systems. Note that the following code is written such that the order of our dynamical variables does not matter.

We implement the differential equation as a dictionary, mapping each of the dynamical variables to its derivative:

.. literalinclude:: ../examples/lotka_volterra.py
	:start-after: example-st\u0061rt
	:lines: 11-14
	:dedent: 1

Note that there are other ways to define the derivative, e.g., used in the previous and following example.

Now, we have everything, we need to instantiate `jitcode`, i.e., to initialise the integration object:

.. literalinclude:: ../examples/lotka_volterra.py
	:start-after: example-st\u0061rt
	:lines: 16
	:dedent: 1

Before we start integrating, we have to choose an integrator and define initial conditions. We here choose the 5th-order Dormand–Prince method. Moreover the initial :math:`B` shall be 0.5, the initial :math:`R` shall be 0.2, and the starting time should be :math:`t=0`.

.. literalinclude:: ../examples/lotka_volterra.py
	:start-after: example-st\u0061rt
	:lines: 17-19
	:dedent: 1

We then define an array of time points where we want to observe the system, and a dictionary of empty lists that shall be filled with our results.

.. literalinclude:: ../examples/lotka_volterra.py
	:start-after: example-st\u0061rt
	:lines: 21-22
	:dedent: 1

Finally, we perform the actual integration by calling `ODE.integrate` for each of our `times`. After this, we use `ODE.y_dict` to access the current state as a dictionary and appending its values to the respective lists.

.. literalinclude:: ../examples/lotka_volterra.py
	:start-after: example-st\u0061rt
	:lines: 23-26
	:dedent: 1

We can now plot or analyse our data, but that’s beyond the scope of JiTCODE. So we just save it:

.. literalinclude:: ../examples/lotka_volterra.py
	:start-after: example-st\u0061rt
	:lines: 28
	:dedent: 1

Taking everything together, our code is:

.. literalinclude:: ../examples/lotka_volterra.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:linenos:
"""

if __name__ == "__main__":
	# example-start
	from jitcode import y, jitcode
	import numpy as np
	
	γ = 0.6
	φ = 1.0
	ω = 0.5
	ν = 0.5
	
	R,B = dynvars = [ y(i) for i in range(2) ]
	
	lotka_volterra_diff = {
			B:  γ*B - φ*R*B,
			R: -ω*R + ν*R*B,
		}
	
	ODE = jitcode(lotka_volterra_diff)
	ODE.set_integrator("dopri5")
	initial_state = { R: 0.2, B: 0.5 }
	ODE.set_initial_value(initial_state,0.0)
	
	times = np.arange(0.0,100,0.1)
	values = { R: [], B: [] }
	for time in times:
		ODE.integrate(time)
		for dynvar in dynvars:
			values[dynvar].append(ODE.y_dict[dynvar])
	
	np.savetxt("timeseries.dat",np.vstack((values[R],values[B])).T)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
For instance, let’s interpret the system from `example` as two oscillators (which is what it is), one consisting of the first and second and one of the third and fourth component. Furthermore, let’s change the control parameters a bit to make the two oscillators identical. We can then calculate the transversal Lyapunov exponents to the synchronisation manifold as follows (important changes are highlighted):

.. literalinclude:: ../examples/double_fhn_transversal_lyap.py
	:emphasize-lines: 6, 12, 14, 17, 19-20
	:start-after: example-st\u0061rt
	:dedent: 1
	:linenos:

Note that the initial state (line 17) is reduced in dimensionality and there is only one component for each synchronisation group.
"""

if __name__ == "__main__":
	# example-start
	from jitcode import jitcode_transversal_lyap, y
	from scipy.stats import sem
	import numpy as np
	
	a = -0.025794
	b =  0.01
	c =  0.02
	k =  0.128
	
	f = [
		y(0) * ( a-y(0) ) * ( y(0)-1.0 ) - y(1) + k * (y(2) - y(0)),
		b*y(0) - c*y(1),
		y(2) * ( a-y(2) ) * ( y(2)-1.0 ) - y(3) + k * (y(0) - y(2)),
		b*y(2) - c*y(3)
		]
	
	initial_state = np.array([1.,2.])
	
	groups = [ [0,2], [1,3] ]
	ODE = jitcode_transversal_lyap(f, groups=groups)
	ODE.set_integrator("lsoda")
	ODE.set_initial_value(initial_state,0.0)
	
	times = range(10,100000,10)
	lyaps = []
	for time in times:
		lyaps.append(ODE.integrate(time)[1])
	
	lyap = np.average(lyaps[1000:])
	stderr = sem(lyaps[1000:]) # Note that this only an estimate
	print("transversal Lyapunov exponent: % .4f ± %.4f" % (lyap,stderr))




#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Integration test of jitcode_restricted_lyap and jitcode_transversal_lyap by comparing them to each other.
"""

from itertools import combinations
from jitcode import jitcode_restricted_lyap, y, jitcode_transversal_lyap
from jitcxde_common import DEFAULT_COMPILE_ARGS
import numpy as np
from scipy.stats import sem
from symengine import Symbol

a = -0.025794
b =  0.01
c =  0.02
k = Symbol("k")

scenarios = [
	{
	"f":
		[
			y(0)*(a-y(0))*(y(0)-1.0) - y(3) + k*(y(1)-y(0)),
			y(1)*(a-y(1))*(y(1)-1.0) - y(4) + k*(y(2)-y(1)),
			y(2)*(a-y(2))*(y(2)-1.0) - y(5) + k*(y(0)-y(2)),
			b*y(0) - c*y(3),
			b*y(1) - c*y(4),
			b*y(2) - c*y(5),
		],
	"vectors":
		[
			[1.,1.,1.,0.,0.,0.],
			[0.,0.,0.,1.,1.,1.]
		],
	"groups":
		( [0,1,2], [3,4,5] )
	},
	{
	"f":
		[
			y(0)*(a-y(0))*(y(0)-1.0) - y(1) + k*(y(2)-y(0)),
			b*y(0) - c*y(1),
			y(2)*(a-y(2))*(y(2)-1.0) - y(3) + k*(y(4)-y(2)),
			b*y(2) - c*y(3),
			y(4)*(a-y(4))*(y(4)-1.0) - y(5) + k*(y(0)-y(4)),
			b*y(4) - c*y(5),
		],
	"vectors":
		[
			[1.,0.,1.,0.,1.,0.],
			[0.,1.,0.,1.,0.,1.]
		],
	"groups":
		( [0,2,4], [1,3,5] )
	},
]

couplings = [
		{"k":-0.1, "sign": 1},
		{"k":-0.2, "sign": 1},
		{"k": 0.1, "sign":-1},
		{"k": 0.2, "sign":-1},
		{"k": 0  , "sign": 0},
	]

for scenario in scenarios:
	for average_dynamics in [False,True]:
		n = len(scenario["f"])
		
		ODE1 = jitcode_restricted_lyap(
				scenario["f"],
				vectors = scenario["vectors"],
				verbose = False,
				control_pars = [k]
			)
		# Simplification or compiler optimisation would lead to trajectories diverging from the synchronisation manifold due to numerical noise.
		ODE1.generate_f_C(simplify=False)
		ODE1.compile_C( extra_compile_args = DEFAULT_COMPILE_ARGS + ["-O2"] )
		ODE1.set_integrator("dopri5")
		
		ODE2 = jitcode_transversal_lyap(
				scenario["f"],
				groups = scenario["groups"],
				average_dynamics = average_dynamics,
				verbose = False,
				control_pars = [k],
			)
		ODE2.set_integrator("dopri5")
		
		for coupling in couplings:
			ODE1.set_parameters(coupling["k"])
			ODE2.set_parameters(coupling["k"])
			
			if coupling["sign"]<0:
				initial_state = np.random.random(n)
			else:
				single = np.random.random(2)
				initial_state = np.empty(n)
				for j,group in enumerate(scenario["groups"]):
					for i in group:
						initial_state[i] = single[j]
			ODE1.set_initial_value(initial_state,0.0)
			
			ODE2.set_initial_value(np.random.random(2),0.0)
			
			times = range(100,100000,100)
			lyaps1 = np.hstack([ODE1.integrate(time)[1] for time in times])
			lyaps2 = np.hstack([ODE2.integrate(time)[1] for time in times])
			
			# Check that we are still on the synchronisation manifold:
			for group in scenario["groups"]:
				for i,j in combinations(group,2):
					assert ODE1.y[i]==ODE1.y[j], "If this fails, the test is broken, not JiTCODE itself."
			
			Lyap1 = np.average(lyaps1[500:])
			Lyap2 = np.average(lyaps2[500:])
			margin1 = sem(lyaps1[500:])
			margin2 = sem(lyaps2[500:])
			sign1 = np.sign(Lyap1) if abs(Lyap1)>margin1 else 0
			sign2 = np.sign(Lyap2) if abs(Lyap2)>margin2 else 0
			assert sign1==coupling["sign"]
			assert sign2==coupling["sign"]
			assert abs(Lyap1-Lyap2)<max(margin1,margin2), "%f±%f \t %f±%f"%(Lyap1,margin1,Lyap2,margin2)
			print( ".", end="", flush=True )

print("")



#!/usr/bin/python3
# -*- coding: utf-8 -*-

from jitcode import jitcode_restricted_lyap, y
import numpy as np
from scipy.stats import sem

a  = -0.025794
b1 =  0.01
b2 =  0.01
c  =  0.02
k  =  0.128

f = [
	y(0) * ( a-y(0) ) * ( y(0)-1.0 ) - y(1) + k * (y(2) - y(0)),
	b1*y(0) - c*y(1),
	y(2) * ( a-y(2) ) * ( y(2)-1.0 ) - y(3) + k * (y(0) - y(2)),
	b2*y(2) - c*y(3)
	]

initial_state = np.random.random(4)

vectors = [
	np.array([1.,0.,1.,0.]),
	np.array([0.,1.,0.,1.])
	]

ODE = jitcode_restricted_lyap(f, vectors=vectors)
ODE.set_integrator("dopri5")
ODE.set_initial_value(initial_state,0.0)

data = np.hstack([ODE.integrate(T)[1] for T in range(10,100000,10)])

print(np.average(data[500:]), sem(data[500:]))



from ._jitcode import (
		t, y,
		jitcode, jitcode_lyap, jitcode_restricted_lyap, jitcode_transversal_lyap,
		UnsuccessfulIntegration,
		test,
		)

try:
	from .version import version as __version__
except ImportError:
	from warnings import warn
	warn('Failed to find (autogenerated) version.py. Do not worry about this unless you really need to know the version.')



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Suppose our differential equation is :math:`\dot{y} = f(y)` with :math:`y∈ℝ^4`,

.. math::

	f(y) = \\left(
	\\begin{matrix}
	y_0 ( a-y_0 ) ( y_0-1) - y_1 + k (y_2 - y_0) \\\\
	b_1 y_0 - c y_1 \\\\
	y_2 ( a-y_2 ) ( y_2-1 ) - y_3 + k (y_0 - y_2)\\\\
	b_2 y_2 - c y_3
	\\end{matrix} \\right),

and :math:`a = -0.025794`, :math:`b_1 = 0.0065`, :math:`b_2 = 0.0135`, :math:`c = 0.02`, and :math:`k = 0.128`.
Then the following code integrates the above for 100000 time units (after discarding 2000 time units of transients), with :math:`y(t=0) = (1,2,3,4)`, and writes the results to :code:`timeseries.dat`:
"""

if __name__ == "__main__":
	# example-start
	from jitcode import jitcode, y
	import numpy as np
	
	a  = -0.025794
	b1 =  0.0065
	b2 =  0.0135
	c  =  0.02
	k  =  0.128
	
	f = [
		y(0) * ( a-y(0) ) * ( y(0)-1.0 ) - y(1) + k * (y(2) - y(0)),
		b1*y(0) - c*y(1),
		y(2) * ( a-y(2) ) * ( y(2)-1.0 ) - y(3) + k * (y(0) - y(2)),
		b2*y(2) - c*y(3)
		]
	
	initial_state = np.array([1.,2.,3.,4.])
	
	ODE = jitcode(f)
	ODE.set_integrator("dopri5")
	ODE.set_initial_value(initial_state,0.0)
	
	times = 2000+np.arange(100000)
	data = []
	for time in times:
		data.append(ODE.integrate(time))
	
	np.savetxt("timeseries.dat", data)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Tests that the code-generation and compilaton or lambdification, respectively, of the derivative and Jacobian works as intended in all kinds of scenarios.
"""

import unittest

from numpy.random import random
from numpy.testing import assert_allclose

from jitcode import jitcode, jitcode_lyap, y
from jitcode._jitcode import _is_C, _is_lambda

from scenarios import (
		y0, f_of_y0, jac_of_y0,
		vanilla, with_params, with_helpers, with_generator, with_dictionary, callback,
		n, params_args
	)

class TestBasic(unittest.TestCase):
	init_params = ()
	
	def setUp(self):
		self.ODE = jitcode(**vanilla)
	
	def test_default(self):
		self.ODE.generate_f_C()
		self.ODE.generate_jac_C()
		self.ODE.compile_C()
		if isinstance(self.ODE,jitcode) and self.ODE.f_sym():
			self.ODE.check()
		self.assertTrue(_is_C(self.ODE.f))
		self.assertTrue(_is_C(self.ODE.jac))
	
	def test_heavily_chunked(self):
		self.ODE.generate_f_C(chunk_size=1,do_cse=True)
		self.ODE.generate_jac_C(chunk_size=1,do_cse=True)
		self.ODE.generate_helpers_C(chunk_size=1)
		self.ODE.compile_C()
		self.assertTrue(_is_C(self.ODE.f))
		self.assertTrue(_is_C(self.ODE.jac))
	
	def test_omp(self):
		self.ODE.generate_f_C(chunk_size=1,do_cse=True)
		self.ODE.generate_jac_C(chunk_size=1,do_cse=True)
		self.ODE.generate_helpers_C(chunk_size=1)
		self.ODE.compile_C(omp=True)
		self.assertTrue(_is_C(self.ODE.f))
		self.assertTrue(_is_C(self.ODE.jac))
	
	def test_lambdas_with_jac(self):
		self.ODE.generate_f_lambda()
		self.ODE.generate_jac_lambda()
		self.assertTrue(_is_lambda(self.ODE.f))
		self.assertTrue(_is_lambda(self.ODE.jac))
	
	def tearDown(self):
		self.assertIsNotNone(self.ODE.f)
		self.ODE.set_parameters(*self.init_params)
		assert_allclose( self.ODE.f(0.0,y0), f_of_y0, rtol=1e-5 )
		if not self.ODE.jac is None:
			assert_allclose( self.ODE.jac(0.0,y0), jac_of_y0, rtol=1e-5)

class TestHelpers(TestBasic):
	def setUp(self):
		self.ODE = jitcode(**with_helpers)

class FurtherHelpersTests(unittest.TestCase):
	def test_identity_of_jacs(self):
		x = random(n)
		
		def evaluate(scenario):
			ODE = jitcode(**scenario)
			ODE.generate_jac_C()
			ODE.compile_C()
			return ODE.jac(0.0,x)
		
		ODE2 = jitcode(**with_helpers)
		ODE2.generate_jac_C()
		ODE2.compile_C()
		
		assert_allclose(
				evaluate(vanilla),
				evaluate(with_helpers)
			)
	
	def test_identity_of_lyaps(self):
		x = random((n+1)*n)
		
		def evaluate(scenario):
			ODE = jitcode_lyap(**scenario,n_lyap=n)
			ODE.compile_C()
			return ODE.f(0.0,x)
		
		assert_allclose(
				evaluate(vanilla),
				evaluate(with_helpers)
			)

class TestGenerator(TestBasic):
	def setUp(self):
		self.ODE = jitcode(**with_generator)

class TestDictionary(TestBasic):
	def setUp(self):
		self.ODE = jitcode(**with_dictionary)

class TestParams(TestBasic):
	init_params = params_args
	
	def setUp(self):
		self.ODE = jitcode(**with_params)

class TestCallBack(unittest.TestCase):
	def test_callback(self):
		ODE = jitcode(**callback)
		ODE.generate_f_C()
		ODE.compile_C()
		ODE.check()
		self.assertTrue(_is_C(ODE.f))
		ODE._initialise(*[fun for _,fun,_ in callback["callback_functions"]])
		assert_allclose( ODE.f(0.0,y0), f_of_y0, rtol=1e-5 )
	
	def test_lambdas(self):
		ODE = jitcode(**callback)
		with self.assertRaises(NotImplementedError):
			ODE.generate_f_lambda()

if __name__ == "__main__":
	unittest.main(buffer=True)




import sympy

#: the symbol for the state that must be used to define the differential equation. It is a function and the integer argument denotes the component. This one is different from the one you can import from jitcode directly by being defined via SymPy and thus being better suited for some symbolic processing techniques that are not available in SymEngine yet.
y = sympy.Function("y",real=True)

#: the symbol for time for defining the differential equation. This one is different from the one you can import from jitcode directly by being defined via SymPy and thus being better suited for some symbolic processing techniques that are not available in SymEngine yet.
t = sympy.Symbol("t", real=True)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
This example showcases several advanced features of JiTCODE that are relevant for an efficient integration of more complex systems as well as how to deal with some special situations. Therefore it is pretty bizarre from a dynamical perspective.

Suppose we want to integrate a system of :math:`N=500` Rössler oscillators, with the :math:`i`-th oscillator being described by the following differential equations (note that we used :math:`v` instead of :math:`y` for the second dynamical variable to avoid overloading symbols):

.. math::
	\\begin{alignedat}{1}
	\\dot{x}_i &= -ω_i v_i - z_i + k \\sin(t) \\sum_{j=0}^N A_{ij} (x_j-x_i) \\\\
	\\dot{v}_i &= ω_i x_i + a v_i \\\\
	\\dot{z}_i &= b + z_i (x_i -c) + k \\sum_{j=0}^N (z_j-z_i)
	\\end{alignedat}

The control parameters shall be :math:`a = 0.165`, :math:`b = 0.2`, :math:`c = 10.0`, and :math:`k = 0.01`. The (frequency) parameter :math:`ω_i` shall be picked randomly from the uniform distribution on :math:`[0.8,1.0]` for each :math:`i`. :math:`A∈ℝ^{N×N}` shall be the adjacency matrix of a one-dimensional small-world network (which shall be provided by a function `small_world_network` in the following example code). So, the :math:`x` compenents are coupled diffusively with a small-world coupling topology, while the :math:`z` components are coupled diffusively to their mean field, with the coupling term being modulated with :math:`\\sin(t)`.

Without further ado, here is the example code (`complete running example <https://raw.githubusercontent.com/neurophysik/jitcode/master/examples/SW_of_Roesslers.py>`_); highlighted lines will be commented upon below:

.. literalinclude:: ../examples/SW_of_Roesslers.py
	:linenos:
	:dedent: 1
	:lines: 63-
	:emphasize-lines: 9, 25-26, 34, 31, 43

Explanation of selected features and choices:

* The values of :math:`ω` are initialised globally (line 9). We cannot just define a function here, because the parameter is used twice for each oscillator. Moreover, if we were trying to calculate Lyapunov exponents or the Jacobian, the generator function would be called multiple times, and thus the value of the parameter would not be consistent (which would be desastrous).

* Since we need :math:`\\sum_{j=0}^N x_j` to calculate the derivative of :math:`z` for every oscillator, it is prudent to only calculate this once. Therefore we define a helper symbol for this in lines 25 and 26, which we employ in line 34. (See the arguments of `jitcode` for details.)

* In line 31, we implement :math:`\\sin(t)`. For this purpose we had to import `t` in line 1. Also, we need to use `symengine.sin` – in contrast to `math.sin` or `numpy.sin`.

* As this is a large system, we use a generator function instead of a list to define :math:`f` (lines 28-35) and have the code automatically be split into chunks of 150 lines, corresponding to the equations of fifty oscillators, in line 43. (For this system, any reasonably sized multiple of 3 is a good choice for `chunk_size`; for other systems, the precise choice of the value may be more relevant.) See `large_systems` for the rationale.
"""

if __name__ == "__main__":
	def small_world_network(number_of_nodes, nearest_neighbours, rewiring_probability):
		n = number_of_nodes
		m = nearest_neighbours//2
		
		A = np.zeros( (n,n), dtype=bool )
		for i in range(n):
			for j in range(-m,m+1):
				A[i,(i+j)%n] = True
		
		# rewiring
		for i in range(n):
			for j in range(i):
				if A[i,j] and (np.random.random() < rewiring_probability):
					A[j,i] = A[i,j] = False
					while True:
						i_new,j_new = np.random.randint(0,n,2)
						if A[i_new,j_new] or i_new==j_new:
							continue
						else:
							A[j_new,i_new] = A[i_new,j_new] = True
							break
		
		return A
	
	# example-start
	from jitcode import jitcode, y, t
	import numpy as np
	import symengine
	
	# parameters
	# ----------
	
	N = 500
	ω = np.random.uniform(0.8,1.0,N)
	a = 0.165
	b = 0.2
	c = 10.0
	k = 0.01
	
	# get adjacency matrix of a small-world network
	A = small_world_network(
		number_of_nodes = N,
		nearest_neighbours = 20,
		rewiring_probability = 0.1
		)
	
	# generate differential equations
	# -------------------------------
	
	sum_z = symengine.Symbol("sum_z")
	helpers = [( sum_z, sum( y(3*j+2) for j in range(N) ) )]
	
	def f():
		for i in range(N):
			coupling_sum = sum( y(3*j)-y(3*i) for j in range(N) if A[i,j] )
			coupling_term = k * symengine.sin(t) * coupling_sum
			yield -ω[i] * y(3*i+1) - y(3*i+2) + coupling_term
			yield  ω[i] * y(3*i) + a*y(3*i+1)
			coupling_term_2 = k * (sum_z-N*y(3*i+2))
			yield b + y(3*i+2) * (y(3*i) - c) + coupling_term_2
	
	# integrate
	# ---------
	
	initial_state = np.random.random(3*N)
	
	ODE = jitcode(f, helpers=helpers, n=3*N)
	ODE.generate_f_C(simplify=False, do_cse=False, chunk_size=150)
	ODE.set_integrator('dopri5')
	ODE.set_initial_value(initial_state,0.0)
	
	# data structure: x[0], v[0], z[0], x[1], …, x[N], v[N], z[N]
	data = np.vstack([ODE.integrate(T) for T in range(10,100000,10)])



from inspect import signature

from scipy.integrate import ode
from scipy.integrate._ode import find_integrator
from scipy.integrate._ivp.ivp import METHODS as ivp_methods
from numpy import inf

class UnsuccessfulIntegration(Exception):
	"""
		This exception is raised when the integrator cannot meet the accuracy and step-size requirements. If you want to know the exact state of your system before the integration fails or similar, catch this exception.
	"""
	pass

def integrator_info(name):
	"""
	Finds out the integrator from a given name, what backend it uses, and whether it can use a Jacobian.
	"""
	if name == 'zvode':
		raise NotImplementedError("JiTCODE does not natively support complex numbers yet.")
	
	if name in ivp_methods.keys():
		integrator = ivp_methods[name]
		return {
				"backend": "ivp",
				"wants_jac": "jac" in signature(integrator).parameters,
				"integrator": integrator
			}
	else:
		integrator = find_integrator(name)
		if integrator is None:
			raise RuntimeError("There is no integrator with that name; using fallback.")
		return {
				"backend": "ode",
				"wants_jac": "with_jacobian" in signature(integrator).parameters,
				"integrator": integrator
			}

class IVP_wrapper(object):
	"""
	This is a wrapper around the integrators from scipy.integrate.solve_ivp making them work like scipy.integrate.ode (mostly).
	"""
	
	def __init__(
			self,
			name,
			f, jac=None,
			initialiser = lambda:None,
			**kwargs
		):
		info = integrator_info(name)
		self.ivp_class = info["integrator"]
		self.f = f
		self.jac = jac
		self.wants_jac = info["wants_jac"]
		self.initialiser = initialiser
		
		# Dictionary to be passed as arguments to the integrator and store stuff
		self.kwargs = {
				"t_bound": inf,
				"vectorized": False,
				"fun": self.f
			}
		if self.wants_jac:
			self.kwargs["jac"] = self.jac
		self.kwargs.update(kwargs)
	
	def set_integrator(self,*args,**kwargs):
		raise AssertionError("This method should not be called")
	
	@property
	def _y(self):
		return self.kwargs["y0"]
	
	@property
	def t(self):
		return self.kwargs["t0"]
	
	def try_to_initiate(self):
		"""
		initiate the integrator if all required arguments have been set
		"""
		if (
				"t0" in self.kwargs and
				"y0" in self.kwargs
			):
			self.initialiser()
			self.backend = self.ivp_class(**self.kwargs)
	
	def set_initial_value(self, initial_value, time=0.0):
		self.kwargs["t0"] = time
		self.kwargs["y0"] = initial_value
		self.try_to_initiate()
	
	def set_params(self,*args):
		raise NotImplementedError("This method should not be called anymore")
	
	def integrate(self,t):
		while self.backend.t < t or self.backend.t_old is None:
			self.backend.step()
			if self.backend.status == "failed":
				raise UnsuccessfulIntegration
		self.kwargs["y0"] = self.backend.dense_output()(t)
		self.kwargs["t0"] = t
		return self.kwargs["y0"]
	
	def successful(self):
		return self.backend.status != "failed"

class IVP_wrapper_no_interpolation(IVP_wrapper):
	def integrate(self,t):
		if self.backend.t < t:
			# Manually un-finish the integrator:
			self.backend.t_bound = t
			self.backend.status = "running"
			
			while self.backend.status == "running":
				self.backend.step()
			
			self.kwargs["y0"] = self.backend.y
			self.kwargs["t0"] = t
			if self.backend.status == "failed":
				raise UnsuccessfulIntegration
		
		elif self.backend.t > t:
			raise ValueError("Target time smaller than current time. Cannot integrate backwards in time")
		
		return self.kwargs["y0"]

class ODE_wrapper(ode):
	"""
	This is a wrapper around Scipy’s ODE that mainly avoids confusing and pointless error messages and throws proper exceptions.
	"""
	def integrate(self,t,step=False,relax=False):
		if t>self.t or step or relax:
			result = super(ODE_wrapper,self).integrate(t,step,relax)
			if self.successful():
				return result
			else:
				raise UnsuccessfulIntegration
		elif t==self.t:
			return self._y
		else:
			raise ValueError("Target time smaller than current time. Cannot integrate backwards in time")
	
	@property
	def params(self):
		return self.f_params
	
	def set_params(self,*args):
		raise NotImplementedError("This method should not be called anymore")

class empty_integrator(object):
	"""
	This is a dummy class that mimicks some basic properties of scipy.integrate.ode or the above wrappers, respectively. It exists to store states and parameters and to raise exceptions in the same interface.
	"""

	def __init__(self):
		self.params = ()
		self._y = []
		self._t = None
	
	@property
	def t(self):
		if self._t is None:
			raise RuntimeError("You must call set_initial_value first.")
		else:
			return self._t
	
	def set_integrator(self,*args,**kwargs):
		raise RuntimeError("This function must not be called")
	
	def set_initial_value(self, initial_value, time=0.0):
		self._y = initial_value
		self._t = time
	
	def set_params(self,*args):
		raise NotImplementedError("This method should not be called anymore")
	
	def integrate(self,*args,**kwargs):
		raise RuntimeError("You must call set_integrator first.")
	
	def successful(self):
		raise RuntimeError("You must call set_integrator first.")
	



#!/usr/bin/python3
# -*- coding: utf-8 -*-

from warnings import warn
from types import FunctionType, BuiltinFunctionType
from inspect import signature
from itertools import count

from numpy import hstack, log
import numpy as np
import symengine

from jitcxde_common import jitcxde, checker
from jitcxde_common.helpers import sympify_helpers, sort_helpers, find_dependent_helpers
from jitcxde_common.numerical import random_direction, orthonormalise_qr
from jitcxde_common.symbolic import collect_arguments, ordered_subs, replace_function
from jitcxde_common.transversal import GroupHandler

from jitcode.integrator_tools import empty_integrator, IVP_wrapper, IVP_wrapper_no_interpolation, ODE_wrapper, integrator_info, UnsuccessfulIntegration

#: the symbol for the state that must be used to define the differential equation. It is a function and the integer argument denotes the component. You may just as well define an analogous function directly with SymEngine or SymPy, but using this function is the best way to get the most of future versions of JiTCODE, in particular avoiding incompatibilities. You can import a SymPy variant from the submodule `sympy_symbols` instead (see `SymPy vs. SymEngine`_ for details).
y = symengine.Function("y")

#: the symbol for time for defining the differential equation. If your differential equation has no explicit time dependency (“autonomous system”), you do not need this. You may just as well define an analogous symbol directly with SymEngine or SymPy, but using this function is the best way to get the most of future versions of JiTCODE, in particular avoiding incompatibilities. You can import a SymPy variant from the submodule `sympy_symbols` instead (see `SymPy vs. SymEngine`_ for details).
t = symengine.Symbol("t", real=True)

def _is_C(function):
	return isinstance(function, BuiltinFunctionType)

def _is_lambda(function):
	return isinstance(function, FunctionType)

def _jac_from_f_with_helpers(f, helpers, simplify, n):
	dependent_helpers = [
			find_dependent_helpers(helpers,y(i))
			for i in range(n)
		]
	
	def line(f_entry):
		for j in range(n):
			entry = f_entry.diff(y(j))
			for helper in dependent_helpers[j]:
				entry += f_entry.diff(helper[0]) * helper[1]
			if simplify:
				entry = entry.simplify(ratio=1.0)
			yield entry
	
	for f_entry in f():
		yield line(f_entry)

class jitcode(jitcxde):
	"""
	Parameters
	----------
	f_sym : iterable of symbolic expressions or generator function yielding symbolic expressions or dictionary
		If an iterable or generator function, the `i`-th element is the `i`-th component of the value of the ODE’s derivative :math:`f(t,y)`. If a dictionary, it has to map the dynamical variables to its derivatives and the dynamical variables must be `y(0), y(1), …`.
	
	helpers : list of length-two iterables, each containing a symbol and a symbolic expression
		Each helper is a variable that will be calculated before evaluating the derivative and can be used in the latter’s computation. The first component of the tuple is the helper’s symbol as referenced in the derivative or other helpers, the second component describes how to compute it from `t`, `y` and other helpers. This is for example useful to realise a mean-field coupling, where the helper could look like `(mean, sum(y(i) for i in range(100))/100)`. (See `example_2` for an example.)
	
	wants_jacobian : boolean
		Tell JiTCODE to calculate and compile the Jacobian. For vanilla use, you do not need to bother about this as this is automatically set to `True` if the selected method of integration desires the Jacobian. However, it is sometimes useful if you want to manually apply some code-generation steps (e.g., to apply some tweaks).
		
	n : integer
		Length of `f_sym`. While JiTCODE can easily determine this itself (and will, if necessary), this may take some time if `f_sym` is a generator function and `n` is large. Take care that this value is correct – if it isn’t, you will not get a helpful error message.
	
	control_pars : iterable of symbols
		Each symbol corresponds to a control parameter that can be used when defining the equations and set after compilation using `set_parameters` (in the same order as given here). Using this makes sense if you need to do a parameter scan with short integrations for each parameter and you spend a considerable amount of time compiling.
	
	callback_functions : iterable
		Python functions that should be called at integration time (callback) when evaluating the derivative. Each element of the iterable represents one callback function as a tuple containing (in that order):
		
		*	A SymEngine function object used in `f_sym` to represent the function call. If you want to use any JiTCODE features that need the derivative, this must have a properly defined `f_diff` method with the derivative being another callback function (or constant).
		*	The Python function to be called. This function will receive the state array (`y`) as the first argument. All further arguments are whatever you use as arguments of the SymEngine function in `f_sym`. These can be any expression that you might use in the definition of the derivative and contain, e.g., dynamical variables, time, control parameters, and helpers. The only restriction is that the arguments are floats (and not vectors or similar). The return value must also be a float (or something castable to float). It is your responsibility to ensure that this function adheres to these criteria, is deterministic and sufficiently smooth with respect its arguments; expect nasty errors otherwise.
		*	The number of arguments, **excluding** the state array as mandatory first argument. This means if you have a variadic Python function, you cannot just call it with different numbers of arguments in `f_sym`, but you have to define separate callbacks for each of numer of arguments.
		
		See `this example <https://github.com/neurophysik/jitcdde/blob/master/examples/sunflower_callback.py>`_ (for JiTCDDE) for how to use this.
	
	verbose : boolean
		Whether JiTCODE shall give progress reports on the processing steps.
	
	module_location : string
		location of a module file from which functions are to be loaded (see `save_compiled`). If you use this, you need not give `f_sym` as an argument, but in this case you must give `n`. If you used `control_pars` or `callback_functions`, you have to provide them again. Depending on the arguments you provide, functionalities such as recompiling may not be available; but then the entire point of this option is to avoid these.
	"""
	
	dynvar = y
	
	def __init__(self,
				f_sym = (), *,
				helpers = None,
				wants_jacobian = False,
				n = None,
				control_pars = (),
				callback_functions = (),
				verbose = True,
				module_location = None,
			):
		jitcxde.__init__(self,n,verbose,module_location)
		
		self.f_sym = self._handle_input(f_sym)
		self._f_C_source = False
		self._jac_C_source = False
		self._helper_C_source = False
		self.helpers = sort_helpers(sympify_helpers(helpers or []))
		self.control_pars = control_pars
		self.control_par_values = ()
		self.callback_functions = callback_functions
		self._wants_jacobian = wants_jacobian
		
		self.integrator = empty_integrator()
		
		if self.jitced is None:
			self._initialise = None
			self.f = None
			self.jac = None
		else:
			# Load derivative and Jacobian if a compiled module has been provided
			self._initialise = self.jitced.initialise
			self.f = self.jitced.f
			self.jac = self.jitced.jac if hasattr(self.jitced,"jac") else None
		
		self._number_of_jac_helpers = None
		self._number_of_f_helpers = None
		
		self.general_subs = {
				control_par: symengine.Symbol("parameter_"+control_par.name)
				for control_par in self.control_pars
			}
	
	@checker
	def _check_non_empty(self):
		self._check_assert( self.f_sym(), "f_sym is empty." )
	
	@checker
	def _check_valid_arguments(self):
		for i,entry in enumerate(self.f_sym()):
			for argument in collect_arguments(entry,y):
				self._check_assert(
						argument[0] >= 0,
						"y is called with a negative argument (%i) in equation %i." % (argument[0],i),
					)
				self._check_assert(
						argument[0] < self.n,
						"y is called with an argument (%i) higher than the system’s dimension (%i) in equation %i." % (argument[0],self.n,i)
					)
	
	@checker
	def _check_valid_symbols(self):
		valid_symbols = [t] + [helper[0] for helper in self.helpers] + list(self.control_pars)
		
		for i,entry in enumerate(self.f_sym()):
			for symbol in entry.atoms(symengine.Symbol):
				self._check_assert(
						symbol in valid_symbols,
						"Invalid symbol (%s) in equation %i."  % (symbol.name,i),
					)
	
	@property
	def jac_sym(self):
		if not hasattr(self,"_jac_sym"):
			self.generate_jac_sym()
			self.report("generated symbolic Jacobian")
		return self._jac_sym
	
	def generate_jac_sym(self, simplify=True):
		"""
		generates the Jacobian using SymEngine’s differentiation.
		
		Parameters
		----------
		simplify : boolean
			Whether the resulting Jacobian should be `simplified <http://docs.sympy.org/dev/modules/simplify/simplify.html>`_ (with `ratio=1.0`). This is almost always a good thing.
		"""
		self._jac_sym = _jac_from_f_with_helpers(self.f_sym, self.helpers, simplify, self.n)
	
	def _default_arguments(self):
		basics = [
				("t", "double const"),
				("Y", "PyArrayObject *__restrict const")
			]
		return basics
	
	def _generate_f_C(self):
		if not self._f_C_source:
			self.generate_f_C()
			self.report("generated C code for f")
	
	def generate_f_C(self, simplify=None, do_cse=False, chunk_size=100):
		"""
		translates the derivative to C code using SymEngine’s `C-code printer <https://github.com/symengine/symengine/pull/1054>`_.
		
		Parameters
		----------
		simplify : boolean or None
			Whether the derivative should be `simplified <http://docs.sympy.org/dev/modules/simplify/simplify.html>`_ (with `ratio=1.0`) before translating to C code. The main reason why you could want to enable this is if you expect your derivative not to be optimised and not be so large that simplifying takes a considerable amount of time. If `None`, this will be automatically disabled for `n>10`.
		
		do_cse : boolean
			Whether SymPy’s `common-subexpression detection <http://docs.sympy.org/dev/modules/rewriting.html#module-sympy.simplify.cse_main>`_ should be applied before translating to C code. It is almost always better to let the compiler do this (unless you want to set the compiler optimisation to `-O2` or lower): For simple differential equations this should not make any difference to the compiler’s optimisations. For large ones, it may make a difference but also take long. As this requires all entries of `f` at once, it may void advantages gained from using generator functions as an input. Also, this feature uses SymPy and not SymEngine.
		
		chunk_size : integer
			If the number of instructions in the final C code exceeds this number, it will be split into chunks of this size. See `Handling very large differential equations <http://jitcde-common.readthedocs.io/#handling-very-large-differential-equations>`_ on why this is useful and how to best choose this value.
			If smaller than 1, no chunking will happen.
		"""
		
		self._generate_helpers_C()
		
		# working copy
		f_sym_wc = (entry.subs(self.general_subs) for entry in self.f_sym())
		
		if simplify is None:
			simplify = self.n<=10
		if simplify:
			f_sym_wc = (entry.simplify(ratio=1) for entry in f_sym_wc)
		
		arguments = self._default_arguments()
		
		if self.helpers:
			arguments.append(("general_helper","double const *__restrict const"))
		
		if do_cse:
			import sympy
			get_helper = sympy.Function("get_f_helper")
			set_helper = symengine.Function("set_f_helper")
			
			_cse = sympy.cse(
					sympy.Matrix(sympy.sympify(list(f_sym_wc))),
					symbols = (get_helper(i) for i in count())
				)
			more_helpers = symengine.sympify(_cse[0])
			f_sym_wc = symengine.sympify(_cse[1][0])
			
			if more_helpers:
				arguments.append(("f_helper","double *__restrict const"))
				self.render_and_write_code(
					(set_helper(i,helper[1]) for i,helper in enumerate(more_helpers)),
					name = "f_helpers",
					chunk_size = chunk_size,
					arguments = arguments,
					omp = False,
					)
				self._number_of_f_helpers = len(more_helpers)
		
		set_dy = symengine.Function("set_dy")
		self.render_and_write_code(
				(set_dy(i,entry) for i,entry in enumerate(f_sym_wc)),
				name = "f",
				chunk_size = chunk_size,
				arguments = arguments+[("dY", "PyArrayObject *__restrict const")]
			)
		
		self._f_C_source = True
	
	def _generate_jac_C(self):
		if self._wants_jacobian and not self._jac_C_source:
			self.generate_jac_C()
			self.report("generated C code for Jacobian")
	
	def generate_jac_C(self, do_cse=False, chunk_size=100, sparse=True):
		"""
		translates the symbolic Jacobian to C code using SymEngine’s `C-code printer <https://github.com/symengine/symengine/pull/1054>`_. If the symbolic Jacobian has not been generated, it generates it by calling `generate_jac_sym`.
		
		Parameters
		----------
		
		do_cse : boolean
			Whether SymPy’s `common-subexpression detection <http://docs.sympy.org/dev/modules/rewriting.html#module-sympy.simplify.cse_main>`_ should be applied before translating to C code. It is almost always better to let the compiler do this (unless you want to set the compiler optimisation to `-O2` or lower): For simple differential equations this should not make any difference to the compiler’s optimisations. For large ones, it may make a difference but also take long. As this requires the entire Jacobian at once, it may void advantages gained from using generator functions as an input. Also, this feature uses SymPy and not SymEngine.
		
		chunk_size : integer
			If the number of instructions in the final C code exceeds this number, it will be split into chunks of this size. See `Handling very large differential equations <http://jitcde-common.readthedocs.io/#handling-very-large-differential-equations>`_ on why this is useful and how to best choose this value.
			If smaller than 1, no chunking will happen.
		
		sparse : boolean
			Whether a sparse Jacobian should be assumed for optimisation. Note that this does not mean that the Jacobian is stored, parsed or handled as a sparse matrix. This kind of optimisation would require `ode` or `solve_ivp` to be able to handle sparse matrices without structure in the sparseness.
		"""
		
		self._generate_helpers_C()
		
		# working copy
		jac_sym_wc = ( (entry.subs(self.general_subs) for entry in line) for line in self.jac_sym )
		self.sparse_jac = sparse
		
		arguments = self._default_arguments()
		if self.helpers:
			arguments.append(("general_helper","double const *__restrict const"))
		
		if do_cse:
			import sympy
			get_helper = sympy.Function("get_jac_helper")
			set_helper = symengine.Function("set_jac_helper")
			jac_sym_wc = sympy.Matrix([
					[ sympy.sympify(entry) for entry in line ]
					for line in jac_sym_wc
				])
			
			_cse = sympy.cse(
					sympy.sympify(jac_sym_wc),
					symbols = (get_helper(i) for i in count())
				)
			more_helpers = symengine.sympify(_cse[0])
			jac_sym_wc = symengine.sympify(_cse[1][0].tolist())
			
			if more_helpers:
				arguments.append(("jac_helper","double *__restrict const"))
				self.render_and_write_code(
						(set_helper(i, helper[1]) for i,helper in enumerate(more_helpers)),
						name = "jac_helpers",
						chunk_size = chunk_size,
						arguments = arguments,
						omp = False
					)
				self._number_of_jac_helpers = len(more_helpers)
		
		set_dfdy = symengine.Function("set_dfdy")
		
		self.render_and_write_code(
				(
					set_dfdy(i,j,entry)
					for i,line in enumerate(jac_sym_wc)
					for j,entry in enumerate(line)
					if ( (entry != 0) or not self.sparse_jac )
				),
				name = "jac",
				chunk_size = chunk_size,
				arguments = arguments+[("dfdY", "PyArrayObject *__restrict const")]
			)
		
		self._jac_C_source = True
	
	def _generate_helpers_C(self):
		if self.helpers and not self._helper_C_source:
			self.generate_helpers_C()
			self.report("generated C code for helpers")
	
	def generate_helpers_C(self, chunk_size=100):
		"""
		translates the helpers to C code using SymEngine’s `C-code printer <https://github.com/symengine/symengine/pull/1054>`_.
		
		Parameters
		----------
		chunk_size : integer
			If the number of instructions in the final C code exceeds this number, it will be split into chunks of this size. See `large_systems` on why this is useful.
			
			If there is an obvious grouping of your helpers, the group size suggests itself for `chunk_size`.
			
			If smaller than 1, no chunking will happen.
		"""
		
		if self.helpers:
			get_helper = symengine.Function("get_general_helper")
			set_helper = symengine.Function("set_general_helper")
			
			for i,helper in enumerate(self.helpers):
				self.general_subs[helper[0]] = get_helper(i)
			self.render_and_write_code(
					(set_helper(i, helper[1].subs(self.general_subs)) for i,helper in enumerate(self.helpers)),
					name = "general_helpers",
					chunk_size = chunk_size,
					arguments = self._default_arguments() + [("general_helper","double *__restrict const")],
					omp = False,
				)
		
		self._helper_C_source = True
	
	def _compile_C(self):
		if (not _is_C(self.f)) or self._lacks_jacobian:
			self.compile_C()
			self.report("compiled C code")
	
	def compile_C(
			self,
			extra_compile_args=None, extra_link_args=None,
			verbose = False,
			modulename = None,
			omp = False,
		):
		"""
		compiles the C code (using `Setuptools <http://pythonhosted.org/setuptools/>`_) and loads the compiled functions. If no C code exists, it is generated by calling `generate_f_C` and `generate_jac_C`.
		For detailed information on the arguments and other ways to tweak the compilation, read `these notes <jitcde-common.readthedocs.io>`_.
		
		Parameters
		----------
		extra_compile_args : iterable of strings
		extra_link_args : iterable of strings
			Arguments to be handed to the C compiler or linker, respectively.
		verbose : boolean
			Whether the compiler commands shall be shown. This is the same as Setuptools’ `verbose` setting.
		modulename : string or `None`
			The name used for the compiled module.
		omp : pair of iterables of strings or boolean
			What compiler arguments shall be used for multiprocessing (using OpenMP). If `True`, they will be selected automatically. If empty or `False`, no compilation for multiprocessing will happen (unless you supply the relevant compiler arguments otherwise).
		"""
		
		self._generate_helpers_C()
		self._generate_f_C()
		self._generate_jac_C()
		
		self._process_modulename(modulename)
		
		self._render_template(
				n = self.n,
				has_Jacobian = self._jac_C_source,
				number_of_f_helpers = self._number_of_f_helpers or 0,
				number_of_jac_helpers = self._number_of_jac_helpers or 0,
				number_of_general_helpers = len(self.helpers),
				sparse_jac = self.sparse_jac if self._jac_C_source else None,
				control_pars = [par.name for par in self.control_pars],
				callbacks = [(fun.name,n_args) for fun,_,n_args in self.callback_functions],
			)
		
		self._compile_and_load(
				verbose,
				extra_compile_args, extra_link_args,
				omp
			)
		
		self.f = self.jitced.f
		if hasattr(self.jitced,"jac"):
			self.jac = self.jitced.jac
		self._initialise = self.jitced.initialise
	
	def _prepare_lambdas(self):
		if self.callback_functions:
			raise NotImplementedError("Callbacks do not work with lambdification. You must use the C backend.")
		
		if not hasattr(self,"_lambda_subs") or not hasattr(self,"_lambda_args"):
			if self.helpers:
				warn("Lambdification handles helpers by plugging them in. This may be very inefficient")
			
			self._lambda_subs = list(reversed(self.helpers))
			self._lambda_args = [t]
			for i in range(self.n):
				symbol = symengine.Symbol("dummy_argument_%i"%i)
				self._lambda_args.append(symbol)
				self._lambda_subs.append((y(i),symbol))
			self._lambda_args.extend(self.control_pars)
	
	def _generate_f_lambda(self):
		if not _is_lambda(self.f):
			self.generate_f_lambda()
			self.report("generated lambdified f")
	
	def generate_f_lambda(self, simplify=None, do_cse=False):
		"""
		translates the symbolic derivative to a function using SymEngines `Lambdify` or `LambdifyCSE`.
		
		Parameters
		----------
		simplify : boolean
			Whether the derivative should be `simplified <http://docs.sympy.org/dev/modules/simplify/simplify.html>`_ (with `ratio=1.0`) before translating to C code. The main reason why you could want to enable this is if you expect your derivative not to be optimised and not be so large that simplifying takes a considerable amount of time. If `None`, this will be automatically disabled for `n>10`.
		
		do_cse : boolean
			Whether a common-subexpression detection, namely `LambdifyCSE`, should be used.
		"""
		
		self._prepare_lambdas()
		
		# working copy
		f_sym_wc = (ordered_subs(entry,self._lambda_subs) for entry in self.f_sym())
		
		if simplify is None:
			simplify = self.n<=10
		if simplify:
			f_sym_wc = (entry.simplify(ratio=1.0) for entry in f_sym_wc)
		
		lambdify = symengine.LambdifyCSE if do_cse else symengine.Lambdify
		core_f = lambdify(self._lambda_args,list(f_sym_wc))
		self.f = lambda t,Y: core_f(np.hstack([t,Y,self.control_par_values]))
		
		self.compile_attempt = False
	
	def _generate_jac_lambda(self):
		if not _is_lambda(self.jac):
			self.generate_jac_lambda()
			self.report("generated lambdified Jacobian")
	
	def generate_jac_lambda(self,do_cse=False):
		"""
		translates the symbolic Jacobain to a function using SymEngines `Lambdify` or `LambdifyCSE`. If the symbolic Jacobian has not been generated, it is generated by calling `generate_jac_sym`.
		
		Parameters
		----------
		do_cse : boolean
			Whether a common-subexpression detection, namely `LambdifyCSE`, should be used.
		"""
		
		self._prepare_lambdas()
		
		jac_matrix = symengine.Matrix([
				[ordered_subs(entry,self._lambda_subs) for entry in line]
				for line in self.jac_sym
			])
		
		lambdify = symengine.LambdifyCSE if do_cse else symengine.Lambdify
		core_jac = lambdify(self._lambda_args,jac_matrix)
		self.jac = lambda t,Y: core_jac(np.hstack([t,Y,self.control_par_values]))
		
		self.compile_attempt = False
	
	def generate_lambdas(self):
		"""
		If they do not already exists, this generates lambdified functions by calling `self.generate_f_lambda()` and, if wanted, `generate_jac_lambda()`.
		"""
		
		self._generate_f_lambda()
		if self._wants_jacobian:
			self._generate_jac_lambda()
		self.compile_attempt = False
	
	@property
	def _lacks_jacobian(self):
		return self._wants_jacobian and self.jac is None
	
	def generate_functions(self):
		"""
		The central function-generating function. Tries to compile the derivative and, if wanted, the Jacobian. If this fails, it generates lambdified functions as a fallback.
		"""
		
		if self.compile_attempt is None or self._lacks_jacobian:
			self._attempt_compilation(reset=False)
		
		if not self.compile_attempt:
			self.generate_lambdas()
	
	@property
	def t(self):
		return self.integrator.t
	
	@property
	def _y(self):
		return self.integrator._y
	
	@property
	def y(self):
		return self.integrator._y
	
	@property
	def y_dict(self):
		"""
		The current state of the system as a dictionary mapping dynamical variables to their current value.
		Note that if you use this often, you may want to use `self.y` instead for efficiency.
		"""
		
		return { self.dynvar(i):self.y[i] for i in range(self.n) }
	
	def set_initial_value(self, initial_value, time=0.0):
		"""
		Same as the analogous function in SciPy’s ODE, except that it also accepts the initial_value in form of a dictionary that maps dynamical variables to their initial value.
		"""
		
		if isinstance(initial_value,dict):
			initial_value = self._list_from_dynvar_dict(
					initial_value,
					"initial value",
					self.n,
				)
		
		if self.n != len(initial_value):
			raise ValueError("The dimension of the initial value does not match the dimension of your differential equations.")
		
		self.integrator.set_initial_value(initial_value, time)
	
	def set_integrator(self,name,nsteps=10**6,interpolate=True,**integrator_params):
		"""
		Analogous to the function in SciPy’s ODE with the same name. This automatically generates the derivative and Jacobian if they do not exist yet and are needed. You can also choose integrators from `scipy.integrate.solve_ivp`.
		
		Parameters
		----------
		name: name of the integrator
			One of the following (or a new method supported by either backend):
			
			* `"dopri5"` – Dormand’s and Prince’s explicit fifth-order method via `ode`
			* `"RK45"` – Dormand’s and Prince’s explicit fifth-order method via `solve_ivp`
			* `"dop853"` – DoP853 (explicit) via `ode`
			* `"DOP853"` – DoP853 (explicit) via `solve_ivp`
			* `"RK23"` – Bogacki’s and Shampine’s explicit third-order method via `solve_ivp`
			* `"BDF"` – Implicit backward-differentiation formula via `solve_ivp`
			* `"lsoda"` – LSODA (implicit) via `ode`
			* `"LSODA"` – LSODA (implicit) via `solve_ivp`
			* `"Radau"` – The implicit Radau method via `solve_ivp`
			* `"vode"` – VODE (implicit) via `ode`
			
			The `solve_ivp` methods are usually slightly faster for large differential equations, but they come with a massive overhead that makes them considerably slower for small differential equations. Implicit solvers are slower than explicit ones, except for stiff problems. If you don’t know what to choose, start with `"dopri5"`.
		
		nsteps: integer
 			Same as the respective parameter of the `ode` solvers, but with a higher default value to avoid annoying errors when getting rid of transients.
		
		interpolate: boolean
			Whether the sampled solutions for `solve_ivp` solvers shall be obtained using interpolation. If your sampling step is small, this may make things faster; otherwise it depends. This may also make the results slightly less accurate.
		
		integrator_params
			Parameters passed to the respective integrator. See its documentation for more.
		"""
		
		info = integrator_info(name)
		self._wants_jacobian |= info["wants_jac"]
		
		old_integrator = self.integrator
		
		self.generate_functions()
		
		if info["backend"] == "ode":
			self.initialise()
			self.integrator = ODE_wrapper(self.f,self.jac)
			self.integrator.set_integrator(
					name,
					nsteps = nsteps,
					**integrator_params
				)
		elif info["backend"] == "ivp":
			if not interpolate and name=="LSODA":
				raise NotImplementedError("LSODA doesn’t work without interpolation.")
			IVP = IVP_wrapper if interpolate else IVP_wrapper_no_interpolation
			self.integrator = IVP(
					name,
					self.f,
					initialiser = lambda: self.initialise(force=True),
					jac = self.jac,
					**integrator_params
				)
		
		# Restore state and time, if applicable:
		try:
			old_y = old_integrator._y
			old_t = old_integrator.t
		except (AttributeError,RuntimeError):
			pass
		else:
			self.set_initial_value(old_y,old_t)
	
	def initialise(self,force=False):
		if self._initialise is not None:
			if len(self.control_par_values)==len(self.control_pars):
				self._initialise(
					*self.control_par_values,
					*[callback for _,callback,_ in self.callback_functions]
				)
			elif force:
				raise RuntimeError("Something needs parameters to be set. Try calling `set_parameters` earlier.")
	
	def set_parameters(self,*args):
		"""
		Same as `set_f_params` and `set_jac_params` for SciPy’s ODE (both sets of parameters are set simultaneuosly, because they should be the same anyway).
		
		The parameters can be passed as different arguments or as a list or other sequence.
		"""
		try:
			self.control_par_values = tuple(args[0])
		except (TypeError,IndexError):
			self.control_par_values = args
		else:
			if len(args)>1:
				raise TypeError("Argument must either be a single sequence or multiple numbers.")
		
		if len(self.control_par_values)!=len(self.control_pars):
			raise ValueError("Number of values does not match number of control parameters.")
		
		self.initialise()
	
	def set_f_params(self, *args):
		warn("This function has been replaced by `set_parameters`")
		self.set_parameters(*args)
	
	def set_jac_params(self, *args):
		warn("This function has been replaced by `set_parameters`")
		self.set_parameters(*args)
	
	def integrate(self,*args,**kwargs):
		return self.integrator.integrate(*args,**kwargs)
	
	def successful(self):
		raise NotImplementedError("JiTCODE doesn’t support ODE’s `successful` because it’s a very archaic and tedious way of handling things. Instead the exception `UnsuccessfulIntegration` is raised in case of an unsuccessful integration. If you want your program to do anything other than aborting in case of an unsuccessful integration, catch `UnsuccessfulIntegration` and act upon it as you deem fit.")

class jitcode_lyap(jitcode):
	"""
	the handling is the same as that for `jitcode` except for:
	
	Parameters
	----------
	n_lyap : integer
		Number of Lyapunov exponents to calculate. If negative or larger than the dimension of the system, all Lyapunov exponents are calculated.
	
	simplify : boolean
		Whether the differential equations for the tangent vector shall be subjected to SymEngine’s `simplify`. Doing so may speed up the time evolution but may slow down the generation of the code (considerably for large differential equations). If `None`, this will be automatically disabled for `n>10`.
	"""
	
	def __init__( self, f_sym=(), n_lyap=-1, simplify=None, **kwargs ):
		self.n_basic = kwargs.pop("n",None)
		
		f_basic = self._handle_input(f_sym,n_basic=True)
		self._n_lyap = n_lyap if (0<=n_lyap<=self.n_basic) else self.n_basic
		if self._n_lyap>10:
			warn("You are about to calculate %i Lyapunov exponents. This is very likely more than you need and may lead to severe difficulties with compilation and integration. Unless you really know what you are doing, consider how many Lyapunov exponents you actually need and set the parameter `n_lyap` accordingly." % self._n_lyap)
		
		if simplify is None:
			simplify = self.n_basic<=10
		
		helpers = sort_helpers(sympify_helpers( kwargs.pop("helpers",[]) ))
		
		def f_lyap():
			yield from f_basic()
			
			for i in range(self._n_lyap):
				for line in _jac_from_f_with_helpers(
						f = f_basic,
						helpers = helpers,
						simplify = False,
						n = self.n_basic
					):
					expression = sum(
							entry * y(k+(i+1)*self.n_basic)
							for k,entry in enumerate(line)
							if entry
						)
					if simplify:
						expression = expression.simplify(ratio=1.0)
					yield expression
		
		super(jitcode_lyap, self).__init__(
				f_lyap,
				helpers = helpers,
				n = self.n_basic*(self._n_lyap+1),
				**kwargs
			)
	
	def set_initial_value(self, y, time=0.0):
		if isinstance(y,dict):
			y = self._list_from_dynvar_dict(y,"initial value",self.n_basic)
		
		new_y = [y]
		for _ in range(self._n_lyap):
			new_y.append(random_direction(self.n_basic))
		
		super(jitcode_lyap, self).set_initial_value(hstack(new_y), time)
	
	def set_integrator(self,name,interpolate=None,**kwargs):
		"""
		Same as for `jitcode` except that `interpolate` defaults to `False` for RK45 and Radau to avoid an accumulation error through the nature of the interpolation algorithm. Using `LSODA` as an integrator is discouraged.
		"""
		if interpolate is None:
			interpolate = name in ["RK45","Radau"]
		if name == "LSODA":
			warn("Using LSODA for Lyapunov exponents is discouraged since interpolation errors may accumulate.")
		super(jitcode_lyap,self).set_integrator(name,interpolate,**kwargs)
	
	def norms(self):
		n = self.n_basic
		tangent_vectors = self._y[n:].reshape(self._n_lyap,n)
		tangent_vectors,norms = orthonormalise_qr(tangent_vectors)
		if not np.all(np.isfinite(norms)):
			warn("Norms of perturbation vectors for Lyapunov exponents out of numerical bounds. You probably waited too long before renormalising and should call integrate with smaller intervals between steps (as renormalisations happen once with every call of integrate).")
		return norms, tangent_vectors
	
	def integrate(self, *args, **kwargs):
		"""
		Like SciPy’s ODE’s `integrate`, except for orthonormalising the tangent vectors and:
		
		Returns
		-------
		y : one-dimensional NumPy array
			The state of the system. Same as the output of `jitcode`’s `integrate` and `ode`’s `integrate`.
		
		lyaps : one-dimensional NumPy array
			The “local” Lyapunov exponents as estimated from the growth or shrinking of the tangent vectors during the integration time of this very `integrate` command, i.e., :math:`\\frac{\\ln (α_i^{(p)})}{s_i}` in the notation of [BGGS80]_
		
		lyap_vectors : list of one-dimensional NumPy arrays
			The Lyapunov vectors (normalised) after integration.
		"""
		
		old_t = self.t
		super(jitcode_lyap, self).integrate(*args, **kwargs)
		delta_t = self.t-old_t
		norms, tangent_vectors = self.norms()
		lyaps = log(norms) / delta_t
		self._y[self.n_basic:] = tangent_vectors.flatten()
		super(jitcode_lyap, self).set_initial_value(self._y, self.t)
		
		return self._y[:self.n_basic], lyaps, tangent_vectors
	
	@property
	def y_dict(self):
		return { self.dynvar(i):self.y[i] for i in range(self.n_basic) }


class jitcode_transversal_lyap(jitcode,GroupHandler):
	"""
	Calculates the largest Lyapunov exponent in orthogonal direction to a predefined synchronisation manifold, i.e. the projection of the tangent vector onto that manifold vanishes. In contrast to `jitcode_restricted_lyap`, this performs some transformations tailored to this specific application that may strongly reduce the number of differential equations and ensure a dynamics on the synchronisation manifold.
	
	The handling is the same as that for `jitcode` except for:
	
	Parameters
	----------
	groups : iterable of iterables of integers
		each group is an iterable of indices that identify dynamical variables that are synchronised on the synchronisation manifold.
	
	simplify : boolean
		Whether the transformed differential equations shall be subjected to SymEngine’s `simplify`. Doing so may speed up the time evolution but may slow down the generation of the code (considerably for large differential equations). If `None`, this will be automatically disabled for `n>10`.
	
	average_dynamics : boolean
		Whether the dynamics of synchronised variables shall be averaged over all variables in a synchronisation group. Otherwise, the dynamics of the first variable of a group is taken as representative (the default). Activating this option may be useful if the investigated synchrony is not complete – otherwise it has no effect. Using this voids the advantages of generator functions and may lead to problems with very large dynamics.
	"""
	
	# Read the accompanying paper to understand the internals of this.
	
	def __init__( self, f_sym=(), groups=(), simplify=None, average_dynamics=False, **kwargs ):
		GroupHandler.__init__(self,groups)
		self.n = kwargs.pop("n",None)
		
		f_basic,extracted = self.extract_main(self._handle_input(f_sym))
		helpers = sort_helpers(sympify_helpers( kwargs.pop("helpers",[]) ))
		
		if simplify is None:
			simplify = self.n<=10
		
		z = symengine.Function("z")
		z_vector = [z(i) for i in range(self.n)]
		tangent_vector = self.back_transform(z_vector)
		
		def tangent_vector_f():
			for line in _jac_from_f_with_helpers(
					f = f_basic,
					helpers = helpers,
					simplify = False,
					n = self.n
				):
				yield sum(
						entry * tangent_vector[k]
						for k,entry in enumerate(line)
						if entry
					)
		
		substitutions = {
				y(i): z(self.map_to_main(i))
				for i in range(self.n)
			}
		
		def finalise(entry):
			entry = entry.subs(substitutions)
			if simplify:
				entry = entry.simplify(ratio=1)
			return replace_function(entry,z,y)
		
		if average_dynamics:
			f_list = list(f_basic())
		
		def f_lyap():
			for entry in self.iterate(tangent_vector_f()):
				if type(entry)==int: # i.e., if main index
					if average_dynamics:
						group = groups[entry]
						yield sum( finalise(f_list[i]) for i in group )/len(group)
					else:
						yield finalise(extracted[self.main_indices[entry]])
				else:
					yield finalise(entry[0]-entry[1])
		
		helpers = ((helper[0],finalise(helper[1])) for helper in helpers)
		
		super(jitcode_transversal_lyap, self).__init__(
				f_lyap,
				helpers = helpers,
				n = self.n,
				**kwargs
			)
	
	def set_initial_value(self, y, time=0.0):
		"""
		Like the analogous function of `jitcode`/`scipy.integrate.ode`, except that only one initial value per group of synchronised components has to be provided (in the same order as the `groups` argument of the constructor).
		"""
		if isinstance(y,dict):
			raise NotImplementedError("Dictionary input is not supported for transversal Lyapunov exponents")
		assert len(y)==len(self.groups), "Initial state too long. Provide only one value per synchronisation group"
		
		new_y = np.empty(self.n)
		new_y[self.main_indices] = y
		new_y[self.tangent_indices] = random_direction(len(self.tangent_indices))
		super(jitcode_transversal_lyap, self).set_initial_value(new_y, time)
	
	def set_integrator(self,name,interpolate=False,**kwargs):
		"""
		Same as for `jitcode` except that `interpolate` defaults to `False` for RK45 and Radau to avoid an accumulation error through the nature of the interpolation algorithm. Using `LSODA` as an integrator is discouraged.
		"""
		if interpolate is None:
			interpolate = name in ["RK45","Radau"]
		if name == "LSODA":
			warn("Using LSODA for Lyapunov exponents is discouraged since interpolation errors may accumulate.")
		super(jitcode_transversal_lyap,self).set_integrator(name,interpolate,**kwargs)
	
	def norm(self):
		tangent_vector = self._y[self.tangent_indices]
		norm = np.linalg.norm(tangent_vector)
		tangent_vector /= norm
		if not np.isfinite(norm):
			warn("Norm of perturbation vector for Lyapunov exponent out of numerical bounds. You probably waited too long before renormalising and should call integrate with smaller intervals between steps (as renormalisations happen once with every call of integrate).")
		self._y[self.tangent_indices] = tangent_vector
		return norm
	
	def integrate(self, *args, **kwargs):
		"""
		Like SciPy’s ODE’s `integrate`, except for orthonormalising the tangent vector and:
		
		Returns
		-------
		y : one-dimensional NumPy array
			The state of the system. Only one initial value per group of synchronised components is returned (in the same order as the `groups` argument of the constructor).
		
		lyaps : one-dimensional NumPy array
			The “local” Lyapunov exponent as estimated from the growth or shrinking of the tangent vector during the integration time of this very `integrate` command, i.e., :math:`\\frac{\\ln (α_i^{(p)})}{s_i}` in the notation of [BGGS80]_
		"""
		
		old_t = self.t
		super(jitcode_transversal_lyap, self).integrate(*args, **kwargs)
		delta_t = self.t-old_t
		norm = self.norm()
		lyap = log(norm) / delta_t
		super(jitcode_transversal_lyap, self).set_initial_value(self._y, self.t)
		
		return self._y[self.main_indices], lyap
	
	@property
	def y_dict(self):
		raise NotImplementedError("Dictionary output is not supported for transversal Lyapunov exponents")


class jitcode_restricted_lyap(jitcode_lyap):
	"""
	Calculates the largest Lyapunov exponent in orthogonal direction to a predefined plane, i.e. the projection of the tangent vector onto that plane vanishes. See `this test <https://github.com/neurophysik/jitcode/blob/master/tests/test_restricted_lyap.py>`_ for an example of usage. The handling is the same as that for `jitcode_lyap` except for:
	
	Parameters
	----------
	vectors : iterable of pairs of NumPy arrays
		A basis of the plane, whose projection shall be removed.
	"""
	
	def __init__(self, f_sym=(), vectors=(), **kwargs):
		kwargs["n_lyap"] = 1
		super(jitcode_restricted_lyap,self).__init__(f_sym,**kwargs)
		self.vectors = [ vector/np.linalg.norm(vector) for vector in vectors ]
	
	def norms(self):
		n = self.n_basic
		tangent_vector = self._y[n:2*n]
		for vector in self.vectors:
			tangent_vector -= np.dot(vector, tangent_vector)*vector
		norm = np.linalg.norm(tangent_vector)
		tangent_vector /= norm
		if not np.isfinite(norm):
			warn("Norm of perturbation vector for Lyapunov exponent out of numerical bounds. You probably waited too long before renormalising and should call integrate with smaller intervals between steps (as renormalisations happen once with every call of integrate).")
		return norm, tangent_vector

def test(omp=True,sympy=True):
	"""
		Runs a quick simulation to test whether:
		
		* a compiler is available and can be interfaced by Setuptools,
		* OMP libraries are available and can be assessed,
		* SymPy is available.
		
		The latter two tests can be deactivated with the respective argument. This is not a full software test but rather a quick sanity check of your installation. If successful, this function just finishes without any message.
	"""
	if sympy:
		import sympy
	ODE = jitcode( [y(1),-y(0)], verbose=False )
	ODE.generate_f_C(chunk_size=1)
	ODE.compile_C(omp=omp)
	ODE.set_integrator("dopri5")
	ODE.set_initial_value([1,2])
	ODE.integrate(0.1)
	



"""
Tests whether things works independent of where symbols are imported from.
"""

import jitcode
import jitcode.sympy_symbols
import sympy
import symengine

symengine_manually = [
		symengine.Symbol("t",real=True),
		symengine.Function("y",real=True),
		symengine.cos,
	]

sympy_manually = [
		sympy.Symbol("t",real=True),
		sympy.Function("y",real=True),
		sympy.cos,
	]

jitcode_provisions = [
		jitcode.t,
		jitcode.y,
		symengine.cos,
	]

jitcode_sympy_provisions = [
		jitcode.sympy_symbols.t,
		jitcode.sympy_symbols.y,
		symengine.cos,
	]

mixed = [
		jitcode.sympy_symbols.t,
		jitcode.y,
		sympy.cos,
	]

results = set()

for t,y,cos in [
			symengine_manually,
			sympy_manually,
			jitcode_provisions,
			jitcode_sympy_provisions,
			mixed
		]:
	ODE = jitcode.jitcode( [cos(t)*y(0)], verbose=False )
	ODE.set_integrator("dopri5")
	ODE.set_initial_value([1.0],0.0)
	
	result = ODE.integrate(10)[0]
	results.add(result)

assert len(results)==1




#!/usr/bin/python3
# -*- coding: utf-8 -*-

import numpy as np
from symengine import symbols, Function
from jitcode import y
from random import shuffle

n = 4

# initial condition
y0 = np.array([ -0.00338158, -0.00223185, 0.01524253, -0.00613449 ])

# solutions
f_of_y0 = np.array([
		0.0045396904008868,
		0.00002265673,
		0.0043665702488807,
		0.000328463955
	])

jac_of_y0 = np.array([
		[-0.1088290163008492, -1.  ,  0.128             ,  0.   ],
		[ 0.0065            , -0.02,  0.                ,  0.   ],
		[ 0.128             ,  0.  , -0.0732042758000427, -1.   ],
		[ 0.                ,  0.  ,  0.0135            , -0.02 ]
	])

y1 = np.array([
		 0.0011789485114731,
		-0.0021947158873226,
		 0.0195744683782066,
		-0.0057801623466600,
	])

lyaps = [0.00710, 0, -0.0513, -0.187]

# vanilla

a  = -0.025794
b1 =  0.0065
b2 =  0.0135
c  =  0.02
k  =  0.128

f = [
		y(0) * ( a-y(0) ) * ( y(0)-1.0 ) - y(1) + k * (y(2) - y(0)),
		b1*y(0) - c*y(1),
		y(2) * ( a-y(2) ) * ( y(2)-1.0 ) - y(3) + k * (y(0) - y(2)),
		b2*y(2) - c*y(3)
	]

vanilla = {"f_sym": f}

# as dictionary
f_dict = { y(i):entry for i,entry in reversed(list(enumerate(f))) }
with_dictionary = {"f_sym": f_dict}

# with generator

def f_generator():
	for entry in f:
		yield entry

with_generator = { "f_sym":f_generator, "n":n }

# with helpers

f1, f2, f3, f4 = symbols("f1, f2, f3, f4")
coupling, first_y, first_y_sq = symbols("coupling, first_y, first_y_sq")
a_alt, b1_alt, b2_alt, c_alt, k_alt = symbols("a_alt, b1_alt, b2_alt, c_alt, k_alt")
f_alt = [ f1, f2, f3, f4 ]

f_alt_helpers = [
	( a_alt , a  ),
	( b1_alt, b1 ),
	( b2_alt, b2 ),
	( c_alt,  c  ),
	( k_alt,  k  ),
	( first_y, y(0) ),
	( first_y_sq, first_y**2 ),
	( coupling, k_alt*(y(2)-first_y) ),
	( f1, a_alt*first_y_sq - a_alt*first_y + coupling - first_y**3 + first_y_sq - y(1) ),
	( f2, b1_alt*first_y - c_alt*y(1)),
	( f3, y(2) * ( a_alt-y(2) ) * ( y(2)-1.0 ) - y(3) - coupling ),
	( f4, b2_alt*y(2) - c_alt*y(3) ),
]
shuffle(f_alt_helpers)

with_helpers = {"f_sym": f_alt, "helpers": f_alt_helpers}

# with parameters

a_par, c_par, k_par = symbols("a_par c_par k_par")
f_params_helpers = [ ( coupling, k_par*(y(2)-y(0)) ) ]
f_params = [
	y(0) * ( a_par-y(0) ) * ( y(0)-1.0 ) - y(1) + coupling,
	b1*y(0) - c_par*y(1),
	y(2) * ( a_par-y(2) ) * ( y(2)-1.0 ) - y(3) - coupling,
	b2*y(2) - c_par*y(3)
	]
params_args = (a,c,k)
n_params = 3

with_params = {
		"f_sym": f_params,
		"helpers": f_params_helpers,
		"control_pars": [a_par, c_par, k_par]
	}

# with callbacks

def first_component(Y):
	return Y[0] * ( a-Y[0] ) * ( Y[0]-1.0 ) - Y[1] + k * (Y[2] - Y[0])

c_times = lambda y,arg: c*arg

call_first_component = Function("call_first_component")
call_c_times = Function("call_c_times")

f_callback = [
		call_first_component(),
		b1*y(0) - call_c_times(y(1)),
		y(2) * ( a-y(2) ) * ( y(2)-1.0 ) - y(3) + k * (y(0) - y(2)),
		b2*y(2) - c*y(3),
	]

callback = {
		"f_sym": f_callback,
		"callback_functions": [
			(call_first_component,first_component,0),
			(call_c_times,c_times,1)
		],
	}





#!/usr/bin/python3
# -*- coding: utf-8 -*-

import unittest

import numpy as np
from numpy.testing import assert_allclose
from numpy.random import random

from jitcode import jitcode
from jitcode.integrator_tools import empty_integrator, IVP_wrapper, IVP_wrapper_no_interpolation, ODE_wrapper, UnsuccessfulIntegration

from scenarios import y0, y1, vanilla, n

# Generating compiled functions

ODE = jitcode(**vanilla,verbose=False,wants_jacobian=True)
ODE.compile_C()
f,jac = ODE.f,ODE.jac

# … time-inverted for certain failure

ODE = jitcode(
		f_sym = [-entry for entry in vanilla["f_sym"]],
		verbose=False,
		wants_jacobian=True
	)
ODE.compile_C()
f_back,jac_back = ODE.f,ODE.jac

# -----------------------------

class TestSkeleton(object):
	"""
	This class exists to be inherited by a test that adds self.initialise to intialise self.integrator.
	"""
	
	def control_result(self):
		result = self.integrator.integrate(1.0)
		assert_allclose( result, y1, rtol=1e-4 )
	
	def test_no_params(self):
		self.initialise(f,jac,rtol=1e-5)
		self.integrator.set_initial_value(y0)
		self.control_result()
	
	def test_initial_twice(self):
		self.initialise(f,jac,rtol=1e-5)
		self.integrator.set_initial_value(random(n))
		self.integrator.set_initial_value(y0)
		self.control_result()
	
	def test_zero_integration(self):
		self.initialise(f,jac)
		initial = random(n)
		self.integrator.set_initial_value(initial)
		assert_allclose(initial,self.integrator.integrate(0))
	
	def test_failed_integration(self):
		self.initialise(f_back,jac_back,atol=1e-12,rtol=1e-12)
		self.integrator.set_initial_value([0,1,2,3],0)
		with self.assertRaises(UnsuccessfulIntegration):
			self.integrator.integrate(1)

class TestSkeletonWithBackwards(TestSkeleton):
	def test_backwards_integration(self):
		self.initialise(f,jac)
		self.integrator.set_initial_value([0,1,2,3],0)
		with self.assertRaises(ValueError):
			self.integrator.integrate(-1)

class TestRK45(unittest.TestCase,TestSkeleton):
	def initialise(self,f,jac,**kwargs):
		self.integrator = IVP_wrapper("RK45",f,**kwargs)

class TestRK23(unittest.TestCase,TestSkeleton):
	def initialise(self,f,jac,**kwargs):
		self.integrator = IVP_wrapper("RK45",f,**kwargs)

class TestRadau(unittest.TestCase,TestSkeleton):
	def initialise(self,f,jac,**kwargs):
		self.integrator = IVP_wrapper("Radau",f,jac,**kwargs)

class TestBDF(unittest.TestCase,TestSkeleton):
	def initialise(self,f,jac,**kwargs):
		self.integrator = IVP_wrapper("BDF",f,jac,**kwargs)

class TestRK45_no_interpolation(unittest.TestCase,TestSkeletonWithBackwards):
	def initialise(self,f,jac,**kwargs):
		self.integrator = IVP_wrapper_no_interpolation("RK45",f,**kwargs)

class TestRK23_no_interpolation(unittest.TestCase,TestSkeletonWithBackwards):
	def initialise(self,f,jac,**kwargs):
		self.integrator = IVP_wrapper_no_interpolation("RK45",f,**kwargs)

class TestRadau_no_interpolation(unittest.TestCase,TestSkeletonWithBackwards):
	def initialise(self,f,jac,**kwargs):
		self.integrator = IVP_wrapper_no_interpolation("Radau",f,jac,**kwargs)

class TestBDF_no_interpolation(unittest.TestCase,TestSkeletonWithBackwards):
	def initialise(self,f,jac,**kwargs):
		self.integrator = IVP_wrapper_no_interpolation("BDF",f,jac,**kwargs)

class TestLSODA(unittest.TestCase,TestSkeleton):
	def initialise(self,f,jac,**kwargs):
		self.integrator = IVP_wrapper("LSODA",f,jac,**kwargs)
	
	# Because this integrator likes to flood the screen with status reports rather than throwing an error
	def test_failed_integration(self):
		pass

class TestDopri5(unittest.TestCase,TestSkeletonWithBackwards):
	def initialise(self,f,jac,**kwargs):
		self.integrator = ODE_wrapper(f)
		self.integrator.set_integrator("dopri5")

class TestDop853(unittest.TestCase,TestSkeletonWithBackwards):
	def initialise(self,f,jac,**kwargs):
		self.integrator = ODE_wrapper(f)
		self.integrator.set_integrator("dop853")

class TestLsoda(unittest.TestCase,TestSkeletonWithBackwards):
	def initialise(self,f,jac,**kwargs):
		self.integrator = ODE_wrapper(f,jac)
		self.integrator.set_integrator("lsoda")
	
	# Because this integrator likes to flood the screen with status reports rather than throwing an error
	def test_failed_integration(self):
		pass

class TestVode(unittest.TestCase,TestSkeletonWithBackwards):
	def initialise(self,f,jac,**kwargs):
		self.integrator = ODE_wrapper(f,jac)
		self.integrator.set_integrator("vode")
	
	# Because this integrator likes to flood the screen with status reports rather than throwing an error
	def test_failed_integration(self):
		pass

class TestDummy(unittest.TestCase):
	def setUp(self):
		self.integrator = empty_integrator()
	
	def test_t(self):
		with self.assertRaises(RuntimeError):
			self.integrator.t
	
	def test_set_integrator(self):
		with self.assertRaises(RuntimeError):
			self.integrator.set_integrator("")
	
	def test_integrate(self):
		with self.assertRaises(RuntimeError):
			self.integrator.integrate(2.3)
	
	def test_set_initial(self):
		initial = random(5)
		self.integrator.set_initial_value(initial,1.2)
		assert np.all( self.integrator._y == initial )
		assert self.integrator.t == 1.2

if __name__ == "__main__":
	unittest.main(buffer=True)




#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
For instance, we can calculate and print the Lyapunov exponents for the system from `example` as follows (changes highlighted):
"""

if __name__ == "__main__":
	# example-start
	from jitcode import jitcode_lyap, y
	from scipy.stats import sem
	import numpy as np
	
	a  = -0.025794
	b1 =  0.0065
	b2 =  0.0135
	c  =  0.02
	k  =  0.128
	
	f = [
		y(0) * ( a-y(0) ) * ( y(0)-1.0 ) - y(1) + k * (y(2) - y(0)),
		b1*y(0) - c*y(1),
		y(2) * ( a-y(2) ) * ( y(2)-1.0 ) - y(3) + k * (y(0) - y(2)),
		b2*y(2) - c*y(3)
		]
	
	initial_state = np.array([1.,2.,3.,4.])
	
	n = len(f)
	ODE = jitcode_lyap(f, n_lyap=n)
	ODE.set_integrator("vode")
	ODE.set_initial_value(initial_state,0.0)
	
	times = range(10,100000,10)
	lyaps = []
	for time in times:
		lyaps.append(ODE.integrate(time)[1])
	
	# converting to NumPy array for easier handling
	lyaps = np.vstack(lyaps)
	
	for i in range(n):
		lyap = np.average(lyaps[1000:,i])
		stderr = sem(lyaps[1000:,i]) # Note that this only an estimate
		print("%i. Lyapunov exponent: % .4f ± %.4f" % (i+1,lyap,stderr))
